use chrono::Utc;
use gtk::prelude::*;
use relm4::{
    component::{AsyncComponent, AsyncComponentParts},
    prelude::*,
    AsyncComponentSender,
};

use crate::{
    database::Database,
    sheet::{I64DateTime, Sheet},
    sheet_dao,
    ui::mcdu::McduOutput,
};

use super::{
    mcdu::McduModel,
    sheet_listing::{SheetListingInput, SheetListingModel},
};

pub struct AppModel {
    database: Database,
    mcdu: Controller<McduModel>,
    sheets_listing: Controller<SheetListingModel>,
}

#[derive(Debug)]
pub enum AppInput {
    SearchStarted(String),
    SheetPressed(Sheet),
}

pub struct AppInitData {
    pub sheets: Vec<Sheet>,
    pub database: Database,
}

#[relm4::component(pub, async)]
impl AsyncComponent for AppModel {
    type Input = AppInput;
    type Output = ();
    type Init = AppInitData;
    type CommandOutput = ();

    view! {
        #[root]
        gtk::Window{
            set_default_width: 300,
            set_default_height: 300,
            set_title: Some("Play music!!!"),
            set_maximized: true,
            gtk::Box {
                set_orientation: gtk::Orientation::Horizontal,
                gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand: true,
                    gtk::ScrolledWindow {
                            model.sheets_listing.widget(),
                            set_vexpand: true,
                            set_hexpand: true,
                    },
                },
                model.mcdu.widget() {
                    set_margin_all: 10,
                },
            }
        }
    }

    async fn init(
        init_data: Self::Init,
        window: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        relm4_icons::initialize_icons();

        let mcdu = McduModel::builder()
            .launch(())
            .forward(sender.input_sender(), |response| match response {
                McduOutput::SearchStarted(query) => AppInput::SearchStarted(query),
            });

        let mut sheets = init_data.sheets;
        sheets.sort_by(|a, b| a.cmp(b).reverse());

        let sheets_listing = SheetListingModel::builder()
            .launch(sheets)
            .forward(sender.input_sender(), |response| {
                AppInput::SheetPressed(response.sheet)
            });

        let model = AppModel {
            database: init_data.database,
            mcdu,
            sheets_listing,
        };

        let widgets = view_output!();

        AsyncComponentParts { model, widgets }
    }

    async fn update(
        &mut self,
        message: Self::Input,
        _sender: AsyncComponentSender<Self>,
        _root: &Self::Root,
    ) {
        match message {
            AppInput::SearchStarted(query) => {
                self.sheets_listing
                    .emit(SheetListingInput::Query(query.clone()));
            }
            AppInput::SheetPressed(sheet) => {
                sheet.open_file();
                // TODO: updating directly does not work
                let mut sheet = sheet;
                sheet.last_opened = I64DateTime(Utc::now());
                sheet_dao::update_sheet_last_opened(&self.database, &sheet)
                    .await
                    .unwrap();
            }
        }
    }
}
